#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 27 15:51:23 2022

Prototyping the Streamline-Upwind Petrov Galerkin scheme (SUPG) in 1D
[Brooks1981: SUPG formulations for convection dominated flows with particular emphasis on the incompressible Navier-Stokes equations]

PDE:   - d/dx (k dc/dx) + u dc/dx = q(x)

Inefficient implementation, vividness comes first

@author: dominik.kern@daad-alumni.de
"""

import numpy as np
import matplotlib.pyplot as plt

# boundary conditions
x0 = 0.0
x1 = 1.0
c0 = 0.0   # left BC
c1 = 1.0   # right BC

# mesh
N = 10   # number of elements
nodes = np.linspace(x0, x1, N+1)
elements = [[k, k+1] for k in range(N)]

# material properties
k = lambda x: 1   # diffusion coefficient (in real coordinates)
u = lambda x: 20   # velocity (in real coordinates)
q = lambda x: 0   # source term (in real coordinates)

# trial and test functions on unit interval   
# X = (x-xc)*2/h   x = X*h/2 + xc   dX/dx = 2/h   dx/dX = h/2
N1 = lambda X: 0.5 - 0.5*X   # trial function
N2 = lambda X: 0.5 + 0.5*X
N1X =  lambda X: -0.5
N2X =  lambda X:  0.5
W1 = lambda X: N1(X) - alpha*0.5   # test function
W2 = lambda X: N2(X) + alpha*0.5
W1X =  lambda X: N1X(X)
W2X =  lambda X: N2X(X)

# Gauss-Legendre quadrature on unit interval
num_int = lambda F: (5/9)*F(-np.sqrt(3/5)) + (8/9)*F(0)  + (5/9)*F(np.sqrt(3/5))  

Kglobal = np.zeros((N+1, N+1))   # global system matrix
Fglobal = np.zeros(N+1)   # global right hand side
Pe_avg = 0
for element in elements:
    n0 = element[0]
    n1 = element[1]
    xc = nodes[n0]/2 + nodes[n1]/2
    h = nodes[n1] - nodes[n0]
    U = u(xc)
    K = k(xc)
    Pe = np.abs(U*h/(2*K))   # element Peclet number
    Pe_avg += Pe
    if Pe>0:
        ao =  1/np.tanh(Pe) - 1/Pe     # optimal upwind
    else:
        ao = 0.0
    alpha = ao   # alpha=0 standard Galerkin, alpha=1 full upwind 

    K11i = lambda X: W1X(X)*k(xc+X*h/2)*N1X(X)*2/h + W1(X)*u(xc+X*h/2)*N1X(X)
    K12i = lambda X: W1X(X)*k(xc+X*h/2)*N2X(X)*2/h + W1(X)*u(xc+X*h/2)*N2X(X)
    K21i = lambda X: W2X(X)*k(xc+X*h/2)*N1X(X)*2/h + W2(X)*u(xc+X*h/2)*N1X(X)
    K22i = lambda X: W2X(X)*k(xc+X*h/2)*N2X(X)*2/h + W2(X)*u(xc+X*h/2)*N2X(X)
    K11 = num_int(K11i)   # could still be integrated analytical
    K12 = num_int(K12i)
    K21 = num_int(K21i)
    K22 = num_int(K22i)
    Kglobal[n0,n0] += K11   # assemble
    Kglobal[n0,n1] += K12
    Kglobal[n1,n0] += K21
    Kglobal[n1,n1] += K22
    
    F1i = lambda X: W1(X)*q(xc+X*h/2)*h/2   # could still be integrated analytical
    F2i = lambda X: W2(X)*q(xc+X*h/2)*h/2
    F1 =  num_int(F1i)
    F2 =  num_int(F2i)
    Fglobal[n0] += F1   # assemble
    Fglobal[n1] += F2

# Boundary Conditions and set up global LES
A = Kglobal[1:-1, 1:-1]
b = Fglobal[1:-1] - c0*Kglobal[1:-1,0] - c1*Kglobal[1:-1,-1]
x = np.linalg.solve(A, b)
c = np.concatenate([ [c0], x, [c1] ])

Pe_avg/=N
plt.plot(nodes, c, 'ko-')
plt.xlabel("x")
plt.ylabel("c")
plt.title("Averaged element Peclet number Pe = {}".format(Pe_avg))
plt.show()
